% This is a simple demonstration of the Multi-task Gaussian process
% classification with the EP approximation.

clear all
close all
warning off all
load('Toydataset_10DPET')
% 10DPET  :  10 Data Points from Each Task (5 DP : "+1" class,
% 5DP : "-1" class from each task), DP : Data points
% Set Dataset,alpha=1:100
alpha=1;

X=ToyRanDS{alpha,1};
targetTrainMTL=ToyRanDS{alpha,2};
targetTrainMC=ToyRanDS{alpha,3};
for i=1:6
    XtestN{i}=Toy_N{alpha,i};
    XtestV{i}=Toy_V{alpha,i};
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[dim,N]  =  size(X);

m         =  max(targetTrainMC);           % Number of tasks and classes
PoE_task  =  length(find(targetTrainMC==m))% Number of points from each task
Ind_task  =  [];                           % Task Pointers
FPoE  =  [];

for i = 1:m
    Ind_task =  [Ind_task i];
    FPoE     =  [FPoE PoE_task];
end

noT = [FPoE;Ind_task];
t   = noT(2,:);

no_tasks = length(t);

DataCov = 'covar_ARD_exp_amp';
TaskCov = 'covar_T_cor';

% Separate data from each task 
% We assume that we have the same number of Data points from each task
num   =  PoE_task/2;% Number of data points from each class on each task

PoolN =  [];
PoolV =  [];

for i=1:m
    T{i}  =  X(:,(i-1)*(2*num)+1:(i-1)*(2*num)+2*num);
    Ty{i} =  targetTrainMTL((i-1)*(2*num)+1:(i-1)*(2*num)+2*num); %Targets from each task

    PoolN=[PoolN T{i}(:,1:num)];
    
    PoolV=[PoolV T{i}(:,num+1:end)];
end

% Set number of hyperparameters for the data covariance function
% dim  : number of features for the input  vector
% dx   : number of hyperparameters given the input  covariance function
% no_m : number of task hyperparameters given the task covariance matrix
% N    : number of data points 

C1 = strcmp(DataCov,'covar_ARD_exp');
C2 = strcmp(DataCov,'covar_ARD_exp_amp');
C3 = strcmp(DataCov,'covar_SE_exp');
T1 = strcmp(TaskCov,'covar_T_cor');
T2 = strcmp(TaskCov,'covar_T_chol');
if C1 == 1
    dx = dim;
end
if C2 == 1
    dx = dim+1;
end
if C3 == 1
    dx = 2;
end

if T1 == 1 
    no_m = no_tasks^2 -no_tasks;
end
if T2 == 1
    no_m = no_tasks^2-(no_tasks^2)/2 +no_tasks/2;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%MTL Training using EP%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialize hyperparameters for MTL
hyper_x  = rand(dx,1)
hyper_t  = no_m^2*randn(no_m,1);

hyper    = [hyper_x;hyper_t];

[hyper,LogMar,NoIt] = minimize...
    (hyper, 'GPEP_MTLFrMF', 20, X, targetTrainMTL,noT,targetTrainMC,DataCov,TaskCov);


% transpose the vector of hyperparameters for the scg function
% MargL_EP_MTLFrMF  : computes the marginal likelihood for MTL with EP
% given the hyperparameters
% Grad_EP_MTLFrMF   : computes the gradients with respect to each
% hyperparameter
opt_vect      = zeros(1, 18);
opt_vect(2:3) = 1e-4;
opt_vect(4)   = 1e-6;
opt_vect(9)   = 0;
opt_vect(16)  = 1e-8;
opt_vect(17)  = 0.1;
% 
% hyper = hyper';
% hyper = scg('MargL_EP_MTLFrMF', hyper,opt_vect,'Grad_EP_MTLFrMF', X, targetTrainMTL,noT,targetTrainMC,DataCov,TaskCov);
% hyper = hyper';

% Separate hyperparameters for the data and task covariance function
hyper_x = hyper(1:dx);
hyper_t = hyper(dx+1:end);

[Kt,L_task] = feval(TaskCov,t,hyper_t);%Compute the Task covariance matrix given the optimized task hyperparameters

% Mode finding for MTL EP approximation
[niHat,tafHat,KGS,Kt,Shat,SHatH,L,zitatest,mi,logZepMTL] = MTL_EP_MF...
    (X,targetTrainMTL,Kt,hyper_x,noT,targetTrainMC,DataCov);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Single-task training using EP (training each task separately)
% MargL_EP_Binary  : computes the marginal likelihood for STL with EP
% Grad_EP_Binary   : computes the gradients with respect to each
% hyperparameter for binary classification
% GPBC_EP_MF       : mode finding for binary classification with EP
for i=1:m
    hyper_STL{i}=rand(dx,1)';% Initialize hyperparameters for each task
    
    hyper_STL{i} = scg('MargL_EP_Binary', hyper_STL{i},...
        opt_vect,'Grad_EP_Binary', T{i}, Ty{i},DataCov);
    
    hyper_STL{i} = hyper_STL{i}';
    
    %Mode finding for Single task learning with the EP approximation
    [niHat_STL{i},tafHat_STL{i},logZep_STL{i},K_STL{i},Shat_STL{i} ...
        ,SHatH_STL{i},L_STL{i},zitatest_STL{i},mi_STL{i}]...
        = GPBC_EP_MF(T{i},Ty{i},hyper_STL{i},DataCov);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Training by pooling all data together
hyper_Pool = rand(dx,1)';
hyper_Pool = scg('MargL_EP_Binary', hyper_Pool,opt_vect,'Grad_EP_Binary', X, targetTrainMTL,DataCov);
hyper_Pool = hyper_Pool';
    
[niHat_Pool,tafHat_Pool,logZep_Pool,K_Pool,Shat_Pool,...
    SHatH_Pool,L_Pool,zitatest_Pool,mi_Pool] =...
    GPBC_EP_MF(X,targetTrainMTL,hyper_Pool,DataCov);
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Decision Boundaries
% test_MTL_EP   : test function for MTL GP classification with EP.
% test_BGPC_EP  : test function for STL GP classification with EP.

xmax  =  max(max(X))+2;
xmin  =  min(min(X))-1;
estr  =  xmin:0.2:xmax;

for i=1:length(estr)
    for j=1:length(estr)
        a=[estr(i);estr(j)];
        
        % MTL test phase
        [ppMTL,vvMTL]=test_MTL_EP(a,X,targetTrainMTL,KGS,...
            Kt,hyper_x,niHat,tafHat,Shat,SHatH,L,zitatest,t ...
            ,targetTrainMC,DataCov);
        
        % Pool test phase
        [PP_Pool(i,j)]=test_BGPC_EP(a,X,targetTrainMTL,...
            K_Pool,niHat_Pool,tafHat_Pool,hyper_Pool,...
            Shat_Pool,SHatH_Pool,L_Pool,zitatest_Pool,DataCov);
        
        for k=1:m
           
            % STL test phase
            [PP_STL{k}(i,j)]=test_BGPC_EP(a,T{k},Ty{k},...
                K_STL{k},niHat_STL{k},tafHat_STL{k},...
                hyper_STL{k},Shat_STL{k},SHatH_STL{k},L_STL{k}...
                ,zitatest_STL{k},DataCov);
        
            % Separation of the post. probabilities of each task from the
            % MTL 
            PPMTL{k}(i,j)=ppMTL(k);
            VVMTL{k}(i,j)=vvMTL(k);
           
        end
    
        
    end
end


% Scatter plot and Decision boundaries for MTL & STL 
for k=1:m    
            
    figure
    subplot(1,2,1)
    contour(estr,estr,PPMTL{k}'),colorbar
    hold on
    plot(T{k}(1,1:num),T{k}(2,1:num),'b+')
    plot(T{k}(1,num+1:end),T{k}(2,num+1:end),'r+')
    title(strcat('MTL Post. Class probabilities of task" ',num2str(k),'".'))
   
    subplot(1,2,2)
    contour(estr,estr,PP_STL{k}'),colorbar
    hold on
    plot(T{k}(1,1:num),T{k}(2,1:num),'b+')
    plot(T{k}(1,num+1:end),T{k}(2,num+1:end),'r+')
    title(strcat('STL Post. Class probabilities of task" ',num2str(k),'".'))
   
end


% Scatter plot and Decision boundaries for Pool training
figure
contour(estr,estr,PP_Pool'),axis tight,colorbar
hold on
plot(PoolN(1,:),PoolN(2,:),'k+')
plot(PoolV(1,:),PoolV(2,:),'c+')
title(strcat('Post. Class probabilities for class "+1", produced from the Pooling training'))

% Hinton graph of the task covariance matrix, illustrating correlations
% between the tasks
figure
hintonw(Kt)
title('Hinton graph of the Task covariance matrix')
xlabel('Task id'),ylabel('Task id')