function [logZep PD_E]=GPEP_MTLFrMF(hyper,X,Y,noT,t_pointer,DataCov,TaskCov)
% function that computes the log-marginal likelihood and the gradients with
% respect to data and task hyperparameters, to be used with Rasmussen's
% minimize.m.

% Implemetation follows the algorithm for the EP approximation described 
% in Gaussian Processes for Machine learning. 

% cumGauss.m , solve_chol.m : functions taken from Rasmussen's implementation
% for GPs to obtain numerical stability.
% sparGPMTL : function that creates the final covariance matrix(Kt \kron Kx)
% after deleting rows and columns for the unseen outputs.

[r_noT,c_noT]  = size(noT);
NN             = noT(1,:); % row vector containing the number of data points in each task
N              = sum(NN);  % number of data points in all tasks
t              = noT(2,:); % task descriptor features to insert into the task cov. function 
no_tasks       = length(t);% Number of tasks
dim            = size(X,1);% Number of features
noise          = 1e-9;     % Set noise for data covariance function


%Set number of hyperparameters for the data covariance function
%Set functions for gradients

C1 = strcmp(DataCov,'covar_ARD_exp');
C2 = strcmp(DataCov,'covar_ARD_exp_amp');
C3 = strcmp(DataCov,'covar_SE_exp');

T1 = strcmp(TaskCov,'covar_T_cor');
T2 = strcmp(TaskCov,'covar_T_chol');
if C1==1
    DataCov_d='covar_ARD_exp_dv';
    dx=dim;
end
if C2==1
    DataCov_d='covar_ARD_exp_amp_dv';
    dx=dim+1;
end
if C3==1
    DataCov_d='covar_SE_exp_dv';
    dx=2;
end

if T1==1
    TaskCov_d='covar_Td_cor';
end
if T2==1
    TaskCov_d='covar_Td_chol';
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hyper_x  =  hyper(1:dx);

hyper_t  =  hyper(dx+1:end);


Kx          = feval(DataCov,X,X,hyper_x)+noise*eye(N);

[Kt,L_task] = feval(TaskCov,t,hyper_t);

KGS         = sparGPMTL(Kt,Kx,t,t_pointer);

niHat    =  zeros(N,1);
    
tafHat   =  zeros(N,1);

tafMinus =  zeros(N,1);

miMinus  =  zeros(N,1);
niMinus  =  zeros(N,1);
mi       =  zeros(1,N);

Sigma    =  KGS;

for kk   = 1:15
        
    for i=1:N    
            
        tafMinus(i) = (1/Sigma(i,i))-tafHat(i);
            
        niMinus(i)  = (mi(i)/Sigma(i,i))-niHat(i);
        
        miMinus(i)  = niMinus(i)/tafMinus(i);
        
        %%Marginal Moments
        
        [m0, m1, m2] = feval('cumGauss', Y(i), niMinus(i)/tafMinus(i), 1/tafMinus(i));
        
        hmu = m1./m0;
        
        hs2 = m2./m0 - hmu^2;         
        
        
        Deltataf(i)  =  (1/hs2)-tafMinus(i)-tafHat(i);
        
        tafHat(i)    =  tafHat(i)+Deltataf(i);
        
        niHat(i)     =  (hmu/hs2)-niMinus(i);
    
        ff           =  1/((1/Deltataf(i))+Sigma(i,i));
        
        gg           =  Sigma(:,i)*Sigma(:,i)';
        
        Sigma        =  Sigma-ff*gg;          
        
        mi           =  Sigma*niHat;
        

    end

    sW  =  sqrt(tafHat);

    B   =  eye(N)+sW*sW'.*KGS;
    
    L   =  chol(B);

    V   =  L'\(repmat(sW,1,N).*KGS);

    Sigma = KGS-V'*V;
    
    mi = Sigma*niHat;

end
TT = ones(length(tafHat),1)+tafHat.*(tafMinus.^-1);      
Fi = cumGauss(Y,miMinus,tafMinus.^-1);

Shat   =  diag(tafHat);

SHatH  =  diag(sW);
logZep =  0.5*sum(log(TT))-...
    sum(log(diag(L)))+...
    0.5*niHat'*(KGS-KGS*SHatH*inv(B)*SHatH*KGS-(inv(diag(tafMinus)+Shat)))*niHat+...
    0.5*miMinus'*diag(tafMinus)*inv(Shat+diag(tafMinus))*(Shat*miMinus-2*niHat)+...
    sum(log(Fi));

logZep = -logZep;

b=niHat-sW.*solve_chol(L,sW.*(KGS*niHat));

Zita=b*b'-repmat(sW,1,N).*solve_chol(L,diag(sW));

for i=1:length(hyper_x)
    
    kxd = feval(DataCov_d,X,X,hyper_x,i);
    
    Cx=sparGPMTL(Kt,kxd,t,t_pointer);
    
    Gradx(i)=-0.5*(trace(Zita*Cx));
    
end

for i=1:length(hyper_t)

    if T1==1 ||T2==1||T3==1
       [dum,dL_task] = feval(TaskCov_d,t,hyper_t,i);
    
       kfd=dL_task*L_task'+L_task*dL_task';
    end
    
    
    Cf=sparGPMTL(kfd,Kx,t,t_pointer);
    
    Gradf(i)=-0.5*(trace(Zita*Cf));
    
end

PD_E=[Gradx';Gradf'];