function [niHat,tafHat,KGS,Kt,Shat,SHatH,L,zitatest,mi,logZep]...
    =MTL_EP_MF(X,Y,Kt,...
    hyper_x,noT,t_pointer,DataCov)

% Mode finding for MTL with EP approximation.
% Implementation follows algorithm described in Gaussian Processes for
% Machine learning

% cumGauss.m , solve_chol.m : functions taken from Rasmussen's implementation
% for GPs to obtain numerical stability.

% sparGPMTL : function that creates the final covariance matrix(Kt \kron Kx)
% after deleting rows and columns for the unseen outputs

iter       = 15;        % number of EP iterations
N          = size(X,2); % number of data points in all tasks
t          = noT(2,:);  % task descriptor features to insert into the task cov. function 
noise      = 1e-9;      % Set noise for data covariance function


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Kx   =  feval(DataCov,X,X,hyper_x)+noise*eye(N);

KGS  =  sparGPMTL(Kt,Kx,t,t_pointer);


niHat    =  zeros(N,1);
tafHat   =  zeros(N,1);
niMinus  =  zeros(N,1);
Deltataf =  zeros(N,1);
tafMinus =  zeros(N,1);
miMinus  =  zeros(N,1);
mi       =  zeros(1,N);
Sigma    =  KGS;
logZep   =  zeros(iter,1);

for kk=1:iter
        
    for i=1:N    
            
        tafMinus(i) = (1/Sigma(i,i))-tafHat(i);
            
        niMinus(i)  = (mi(i)/Sigma(i,i))-niHat(i);
        
        miMinus(i)  = niMinus(i)/tafMinus(i);
        
        %%Marginal Moments
        
        [m0, m1, m2] = feval('cumGauss', Y(i), niMinus(i)/tafMinus(i), 1/tafMinus(i));
        
        hmu = m1./m0;
        
        hs2 = m2./m0 - hmu^2;         
        
        Deltataf(i) = (1/hs2)-tafMinus(i)-tafHat(i);
        
        tafHat(i)   = tafHat(i)+Deltataf(i);
        
        niHat(i)    = (hmu/hs2)-niMinus(i);
    
        ff = 1/((1/Deltataf(i))+Sigma(i,i));
        
        gg = Sigma(:,i)*Sigma(:,i)';
        
        Sigma = Sigma-ff*gg;          
        
        mi = Sigma*niHat;
        
    end

    Shat   =   diag(tafHat);
    
    sW     =   sqrt(tafHat);
    
    B      =   eye(N)+sW*sW'.*KGS;
        
    L      =   chol(B);
    
    V      =   L'\(repmat(sW,1,N).*KGS);
    
    Sigma  =   KGS-V'*V;
    
    mi     =   Sigma*niHat;
    
    TT     =   ones(length(tafHat),1)+tafHat.*(tafMinus.^-1);      

    Fi     =   cumGauss(Y,miMinus,tafMinus.^-1);
    
    SHatH  =   diag(sW);
    
    logZep(kk) =  0.5*sum(log(TT))- sum(log(diag(L)))...
        +0.5*niHat'*(KGS-KGS*SHatH*inv(B)*SHatH*KGS-...
        (inv(diag(tafMinus)+Shat)))*niHat+0.5*miMinus'...
        *diag(tafMinus)*inv(Shat+diag(tafMinus))*...
        (Shat*miMinus-2*niHat)+ sum(log(Fi));
    
end
SHatH=diag(sW);


zitatest=sW.*solve_chol(L,sW.*(KGS*niHat));