function [logZep]=MargL_EP_MTLFrMF(hyper,X,Y,noT,t_pointer,DataCov,TaskCov)
% function that computes the log-marginal likelihood given data and task 
%hyperparameters, to be used with scg.m.

% Implemetation follows the algorithm for the EP approximation described 
% in Gaussian Processes for Machine learning. 

% cumGauss.m , solve_chol.m : functions taken from Rasmussen's implementation
% for GPs to obtain numerical stability.
% sparGPMTL : function that creates the final covariance matrix(Kt \kron Kx)
% after deleting rows and columns for the unseen outputs.

hyper          = hyper';

iter           = 15;        % number of EP iterations
N              = size(X,2);   % number of data points in all tasks
t              = noT(2,:);  % task descriptor features to insert into the task cov. function 
dim            = size(X,1);
noise          = 1e-9;      % Set noise for data covariance function


% Set number of hyperparameters for the data covariance function
% Set functions for gradients

C1 = strcmp(DataCov,'covar_ARD_exp');
C2 = strcmp(DataCov,'covar_ARD_exp_amp');
C3 = strcmp(DataCov,'covar_SE_exp');
C4 = strcmp(DataCov,'covar_NN');
C5 = strcmp(DataCov,'covar_Lin');


if C1==1
    dx=dim;
end
if C2==1
    dx=dim+1;
end
if C3==1
    dx=2;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hyper_x  =  hyper(1:dx);

hyper_t  =  hyper(dx+1:end);


Kx          = feval(DataCov,X,X,hyper_x)+noise*eye(N);

[Kt,L_task] = feval(TaskCov,t,hyper_t);

KGS         = sparGPMTL(Kt,Kx,t,t_pointer);

niHat    =  zeros(N,1);
Deltataf =  zeros(N,1);
tafHat   =  zeros(N,1);
tafMinus =  zeros(N,1);
miMinus  =  zeros(N,1);
niMinus  =  zeros(N,1);
mi       =  zeros(1,N);
Sigma    =  KGS;

for kk   = 1:iter
        
    for i = 1:N    
            
        tafMinus(i) = (1/Sigma(i,i))-tafHat(i);
            
        niMinus(i)  = (mi(i)/Sigma(i,i))-niHat(i);
        
        miMinus(i)  = niMinus(i)/tafMinus(i);
        
        %%Marginal Moments
        
        [m0, m1, m2] = feval('cumGauss', Y(i), niMinus(i)/tafMinus(i), 1/tafMinus(i));
        
        hmu = m1./m0;
        
        hs2 = m2./m0 - hmu^2;         
        
        
        Deltataf(i)  =  (1/hs2)-tafMinus(i)-tafHat(i);
        
        tafHat(i)    =  tafHat(i)+Deltataf(i);
        
        niHat(i)     =  (hmu/hs2)-niMinus(i);
    
        ff           =  1/((1/Deltataf(i))+Sigma(i,i));
        
        gg           =  Sigma(:,i)*Sigma(:,i)';
        
        Sigma        =  Sigma-ff*gg;          
        
        mi           =  Sigma*niHat;
        

    end

    sW  =  sqrt(tafHat);

    B   =  eye(N)+sW*sW'.*KGS;
    
    L   =  chol(B);

    V   =  L'\(repmat(sW,1,N).*KGS);

    Sigma = KGS-V'*V;
    
    mi = Sigma*niHat;

end
TT = ones(length(tafHat),1)+tafHat.*(tafMinus.^-1);      
Fi = cumGauss(Y,miMinus,tafMinus.^-1);

Shat   =  diag(tafHat);

SHatH  =  diag(sW);
logZep =  0.5*sum(log(TT))-...
    sum(log(diag(L)))+...
    0.5*niHat'*(KGS-KGS*SHatH*inv(B)*SHatH*KGS-(inv(diag(tafMinus)+Shat)))*niHat+...
    0.5*miMinus'*diag(tafMinus)*inv(Shat+diag(tafMinus))*(Shat*miMinus-2*niHat)+...
    sum(log(Fi));


logZep = -logZep;
    
