function [PD_E]=Grad_EP_Binary(hyper,X,Y,DataCov)
% function that computes  the gradients with
% respect to  hyperparameters, to be used  with scg.m , for binary
% classification with the EP approximation


hyper    =  hyper';
iter     =  15;
[dim, N] =  size(X);


%Set number of hyperparameters for the data covariance function
%Set functions for gradients
C1 = strcmp(DataCov,'covar_ARD_exp');
C2 = strcmp(DataCov,'covar_ARD_exp_amp');
C3 = strcmp(DataCov,'covar_SE_exp');
C4 = strcmp(DataCov,'covar_NN');
C5 = strcmp(DataCov,'covar_Lin');

if C1==1
    DataCov_d='covar_ARD_exp_dv';
    dx=dim;
end
if C2==1
    DataCov_d='covar_ARD_exp_amp_dv';
    dx=dim+1;
end
if C3==1
    DataCov_d='covar_SE_exp_dv';
    dx=2;
end
if C4==1
    DataCov_d='covar_NN_dv';
    dx=3;
end
if C5==1
    DataCov_d='covar_Lin_dv';
    dx=1;
end


niHat=zeros(N,1);

tafHat=zeros(N,1);

tafMinus=zeros(N,1);

miMinus=zeros(N,1);
niMinus=zeros(N,1);
mi=zeros(1,N);

noise=1e-9;%%Set noise

K=feval(DataCov,X,X,hyper)+noise*eye(N);

Sigma=K;

for k=1:iter
    
    for i=1:N
        
        tafMinus(i)=(1/Sigma(i,i))-tafHat(i);
        
        niMinus(i)=(mi(i)/Sigma(i,i))-niHat(i);
        
        miMinus(i)=niMinus(i)/tafMinus(i);
        
        %%Marginal Moments
        
        [m0, m1, m2] = feval('cumGauss', Y(i), niMinus(i)/tafMinus(i), 1/tafMinus(i));
        
        hmu = m1./m0;
        
        hs2 = m2./m0 - hmu^2;     

        %%Update Site Parameters

        Deltataf(i)=(1/hs2)-tafMinus(i)-tafHat(i);

        tafHat(i)=tafHat(i)+Deltataf(i);
        
        niHat(i)=(hmu/hs2)-niMinus(i);
        
        ff=1/((1/Deltataf(i))+Sigma(i,i));
        
        gg=Sigma(:,i)*Sigma(:,i)';
        
        Sigma=Sigma-ff*gg;      
        
        mi=Sigma*niHat;
        
    end

    Shat  =  diag(tafHat);
    
    sW    =  sqrt(tafHat);
    
    B     =  eye(N)+sW*sW'.*K;
        
    L     =  chol(B);
    
    V     =  L'\(repmat(sW,1,N).*K);
    
    Sigma =  K-V'*V;
    
    mi    =  Sigma*niHat;
    
end

b       =  niHat-sW.*solve_chol(L,sW.*(K*niHat));

Zita    =  b*b'-repmat(sW,1,N).*solve_chol(L,diag(sW));



for l = 1:dx
    
    C        =  feval(DataCov_d,X,X,hyper,l);
    
    PD_E(l)  =  -0.5*(trace(Zita*C));

    
end

 