function [logZep]=MargL_EP_Binary(hyper,X,Y,DataCov)
% function that computes  the log matginal likelihood given
% the  hyperparameters, to be used  with scg.m , for binary
% classification with the EP approximation

hyper   =  hyper';
iter    =  15;
[dim,N] =  size(X);
noise   =  1e-9;    %%Set noise

%Set number of hyperparameters for the data covariance function
%Set functions for gradients
C1 = strcmp(DataCov,'covar_ARD_exp');
C2 = strcmp(DataCov,'covar_ARD_exp_amp');
C3 = strcmp(DataCov,'covar_SE_exp');
C4 = strcmp(DataCov,'covar_NN');
C5 = strcmp(DataCov,'covar_Lin');

if C1==1
    dx=dim;
end
if C2==1
    dx=dim+1;
end
if C3==1
    dx=2;
end
if C4==1
    dx=3;
end
if C5==1
    dx=1;
end



niHat    =   zeros(N,1);

tafHat   =   zeros(N,1);

tafMinus =   zeros(N,1);

miMinus  =   zeros(N,1);

niMinus  =   zeros(N,1);

mi       =   zeros(1,N);

K=feval(DataCov,X,X,hyper)+noise*eye(N);

Sigma=K;

for k=1:iter
    
    for i=1:N
        
        tafMinus(i) = (1/Sigma(i,i))-tafHat(i);
        
        niMinus(i)  = (mi(i)/Sigma(i,i))-niHat(i);
        
        miMinus(i)  = niMinus(i)/tafMinus(i);
        
        %%Marginal Moments
        
        [m0, m1, m2] = feval('cumGauss', Y(i), niMinus(i)/tafMinus(i), 1/tafMinus(i));
        
        hmu = m1./m0;
        
        hs2 = m2./m0 - hmu^2;     

        %%Update Site Parameters

        Deltataf(i)  =  (1/hs2)-tafMinus(i)-tafHat(i);

        tafHat(i)    =  tafHat(i)+Deltataf(i);
        
        niHat(i)     =  (hmu/hs2)-niMinus(i);
        
        ff           =  1/((1/Deltataf(i))+Sigma(i,i));
        
        gg           =  Sigma(:,i)*Sigma(:,i)';
        
        Sigma        =  Sigma-ff*gg;      
        
        mi           =  Sigma*niHat;
        
    end

    Shat =  diag(tafHat);
    
    sW   =  sqrt(tafHat);
    
    B    =  eye(N)+sW*sW'.*K;
        
    L    =  chol(B);
    
    V    =  L'\(repmat(sW,1,N).*K);
    
    Sigma=  K-V'*V;
    
    mi   =  Sigma*niHat;
    
end

SHatH  =  diag(sW);
TT     =  ones(length(tafHat),1)+tafHat.*(tafMinus.^-1);
Fi     =  cumGauss(Y,miMinus,tafMinus.^-1);
logZep =  0.5*sum(log(TT))-sum(log(diag(L)))...
    +0.5*niHat'*(K-K*SHatH*inv(B)*SHatH*K...
    -(inv(diag(tafMinus)+Shat)))*niHat+...
    0.5*miMinus'*diag(tafMinus)*inv(Shat+diag(tafMinus))...
    *(Shat*miMinus-2*niHat)+sum(log(Fi));

logZep =  -logZep;